;; ======================================================================
;; Ladebug functions
;;
;; David LaFrance-Linden (david_ll@zko.dec.com) 14-Sep-99
;; Figure out which gud.el we are using and configure appropriately,
;; taking into account the various changes through the years (1993 to
;; present), trying our best to inter-operate with XEmacs 20.4 which
;; uses a 1993 variant of gud.el.  See also the misc/ladebug.el.history/
;; subdirectory in the ladebug hierarchy.
;;   Testing:
;;	GNU	19.28.1	blkbrd	4.0D /usr/bin/emacs		OK
;;	GNU	19.28.1	ladoga	4.0D /usr/bin/emacs		OK
;;	GNU	19.28.1	wicked	4.0F /usr/bin/emacs		OK
;;	GNU	19.34.1	blkbrd	4.0D /usr/local/bin/emacs 	OK
;;	GNU	19.34.1	ladoga	4.0D /usr/local/bin/emacs 	OK
;;	GNU 19.11 XEmacs achevy T5.0 /usr/local/bin/xemacs	buffsel
;;	XEmacs	19.14	blkbrd  4.0D /usr/local/bin/xemacs	buffsel, prompt
;;	XEmacs	19.14	ladoga  4.0D /usr/local/bin/xemacs	buffsel, bpt
;;	GNU	20.4.1	wicked	4.0F /usr/local/bin/emacs	OK
;;	XEmacs	20.4	achevy	T5.0 /usr/bin/emacs		buffsel
;;	XEmacs	20.4	achevy	T5.0 /usr/bin/xemacs		buffsel, prompt sometimes
;;	GNU	20.3.1	alime	2.0.35 /usr/bin/emacs		OK
;;	GNU	20.3.1	tools	2.2... /usr/bin/emacs		OK
;;    key:
;;	buffsel -- Initial buffer selection glitch -- uses one window
;;		(source) instead of splitting windows (source + *gud*).
;;      bpt -- c-C c-B to set a breakpoint doesn't work (stringp nil)
;;	prompt -- ladebug prompt not coming through
;;
;; Tim Lucia   (lucia@zko.dec.com) 30-Oct-95
;; Modified to work with 19.27 release of GUD and with the ladebug debugger
;; release with Digital UNIX V4.0.  The name was changed from decladebug to
;; just ladebug.
;;
;; Tim Bingham (tjb@zko.dec.com) 28-December-1993
;;
;; Based on a version by Dirk Grunwald (grunwald@foobar.cs.colorado.edu)
;;
;; This version makes use of the -emacs option to Ladebug which causes
;; file/line number markers to be output. These markers have the same syntax
;; as those of gdb and so we inherit gud-gdb-marker-filter instead of defining
;; our own gud-ladebug-marker-filter. Similarly we inherit gud-gdb-find-file
;; rather than define gud-ladebug-find-file.
;;
;; These functions require version 2.0.12 or later of Ladebug
;;

;; We are built in terms of the Grand Unified Debugger
(require 'gud)


;; BEGIN configuration based on gud.el

;; Figure out which emacs gud environment we are in.  This is a real
;; mess, since the interfaces have changed over the years, and XEmacs
;; 20.4 ships (mid'1999) with a VERY ANCIENT gud.el (1993).

(defun arglist-length (fname) 
  (let ((def (and (fboundp fname) (symbol-function fname))))
    (cond ((null def) -1)
	  ((and (fboundp 'compiled-function-p)
		(fboundp 'compiled-function-arglist)
		(compiled-function-p def))
	   (length (compiled-function-arglist def)))
	  ((and (fboundp 'byte-code-function-p)
		(byte-code-function-p def))
	   (length (car (append def nil))))
	  ((eq (car-safe def) 'lambda)
	   (length (nth 1 def)))
	  (t -1))))

(defvar gud-common-init-arglist-length (arglist-length 'gud-common-init))
(defvar gud-uses-overload-techniqe     (fboundp 'gud-overload-functions))

(defvar gud-ladebug-massage-with-file 
  gud-uses-overload-techniqe)

(defvar gud-ladebug-can-use-gdb-marker-filter
  ;; Only the XEmacs variant can't inherit the gdb-marker-filter
  ;; XEmacs apparently being overload technique with 2 common init args
  (not (and gud-uses-overload-techniqe
	    (= gud-common-init-arglist-length 2))))
(defvar gud-ladebug-hack-local-keys
  (= gud-common-init-arglist-length 4))
(defvar gud-ladebug-hack-paragraph-start 
  (not (null (member gud-common-init-arglist-length '(1 4)))))

;; END configuration based on gud.el



;; History of argument lists passed to Ladebug
(defvar gud-ladebug-history nil)

;; ladebug activation name -- changed from decladebug to ladebug
;; in Digital UNIX V4.0.
(defvar gud-ladebug-ladebug-name nil)
(setq gud-ladebug-ladebug-name
      (cond ((file-executable-p "/usr/bin/ladebug") "ladebug ")
	    ((file-executable-p "/usr/bin/decladebug") "decladebug ")
	    (t "ladebug ")))

(defun gud-ladebug-massage-args (file args)
  (cons "-emacs" (if gud-ladebug-massage-with-file
		     (cons file args)
		   args)))



;;BEGIN copied from the 1994 version; we use the pre --annotate=2 method

;; There's no guarantee that Emacs will hand the filter the entire
;; marker at once; it could be broken up across several strings.  We
;; might even receive a big chunk with several markers in it.  If we
;; receive a chunk of text which looks like it might contain the
;; beginning of a marker, we save it here between calls to the
;; filter.
(defvar gud-marker-acc "")
(make-variable-buffer-local 'gud-marker-acc)

(defun gud-ladebug-marker-filter (string)
  (save-match-data
    (setq gud-marker-acc (concat gud-marker-acc string))
    (let ((output ""))

      ;; Process all the complete markers in this chunk.
      (while (string-match "\032\032\\([^:\n]*\\):\\([0-9]*\\):.*\n"
			   gud-marker-acc)
	(setq

	 ;; Extract the frame position from the marker.
	 gud-last-frame
	 (cons (substring gud-marker-acc (match-beginning 1) (match-end 1))
	       (string-to-int (substring gud-marker-acc
					 (match-beginning 2)
					 (match-end 2))))

	 ;; Append any text before the marker to the output we're going
	 ;; to return - we don't include the marker in this text.
	 output (concat output
			(substring gud-marker-acc 0 (match-beginning 0)))

	 ;; Set the accumulator to the remaining text.
	 gud-marker-acc (substring gud-marker-acc (match-end 0))))

      ;; Does the remaining text look like it might end with the
      ;; beginning of another marker?  If it does, then keep it in
      ;; gud-marker-acc until we receive the rest of it.  Since we
      ;; know the full marker regexp above failed, it's pretty simple to
      ;; test for marker starts.
      (if (string-match "\032.*\\'" gud-marker-acc)
	  (progn
	    ;; Everything before the potential marker start can be output.
	    (setq output (concat output (substring gud-marker-acc
						   0 (match-beginning 0))))

	    ;; Everything after, we save, to combine with later input.
	    (setq gud-marker-acc
		  (substring gud-marker-acc (match-beginning 0))))

	(setq output (concat output gud-marker-acc)
	      gud-marker-acc ""))

      output)))

;;END copied from the 1994 version


;;;####autoload
(defun ladebug (command-line)
  "Run ladebug on program FILE in buffer *gud-FILE*.
The directory containing FILE becomes the initial working directory
and source-file directory for your debugger."
  (interactive
   (list (read-from-minibuffer "Run ladebug (like this): "
			       (if (consp gud-ladebug-history)
				   (car gud-ladebug-history)
				 gud-ladebug-ladebug-name)
			       nil nil
			       '(gud-ladebug-history . 1))))

  (cond ((= gud-common-init-arglist-length 4)
	 ;; the most modern interface encountered
	 (gud-common-init command-line
			  'gud-ladebug-massage-args
			  (if gud-ladebug-can-use-gdb-marker-filter
			      'gud-gdb-marker-filter
			    'gud-ladebug-marker-filter)
			  'gud-gdb-find-file))
	(gud-uses-overload-techniqe
	 ;; much earlier technique
	 (gud-overload-functions
	  '((gud-massage-args  . gud-ladebug-massage-args)
	    (gud-marker-filter . gud-ladebug-marker-filter)
	    (gud-find-file     . gud-gdb-find-file)))
	 (cond ((= gud-common-init-arglist-length 1)
		;; early (but developed) GNU emacs interface
		(gud-common-init command-line))
	       ((= gud-common-init-arglist-length 2)
		;; earlier (more prototypical) interface used by "modern" XEmacs (e.g. 20.4)
		(gud-common-init command-line "ladebug"))
	       (t
		(error "gud overload technique, but unknown interface."))))
	(t
	 (error "Can't figure out which 'gud' is being used.")))
  
  (gud-def gud-break  "stop at \"%f\":%l" "\C-b" "Set breakpoint at current line of current buffer.")
  (gud-def gud-step   "step"       "\C-s" "Step one line (into functions).")
  (gud-def gud-stepi  "stepi"      "\C-i" "Step one instruction.")
  (gud-def gud-next   "next"       "\C-n" "Step one line (over functions).")
  (gud-def gud-cont   "cont"	   "\C-r" "Continue.")
  (gud-def gud-finish "return"	   "\C-f" "Return from current function.")
  (gud-def gud-up     "up %p"	   "<"    "Up (numeric arg) stack frames.")
  (gud-def gud-down   "down %p"	   ">"    "Down (numeric arg) stack frames.")
  (gud-def gud-print  "print %e"   "\C-p" "Evaluate expression at point.")

  (if gud-ladebug-hack-local-keys
      (progn 
	(local-set-key [menu-bar debug finish] '("Finish Function" . gud-finish))
	(local-set-key [menu-bar debug up]     '("Up Stack"        . gud-up))
	(local-set-key [menu-bar debug down]   '("Down Stack"      . gud-down))
	))

  (setq comint-prompt-regexp  "^[^)]*ladebug) *")
  (if gud-ladebug-hack-paragraph-start
      (setq paragraph-start comint-prompt-regexp))

  (run-hooks 'ladebug-mode-hook)
  )

(fset 'decladebug 'ladebug)

(provide 'ladebug)

;;; ladebug.el ends here
